#define	FCNTL_H	<fcntl.h>
#define	FREAD_TYPE	int
#define	FWRITE_TYPE	int
#define	GETEUID_TYPE	uid_t
#define	GETUID_TYPE	uid_t
#define	MNTENT_H	<sys/mntent.h>
#define	NDIR_H	<dirent.h>
#define	NDIR_TYPE	dirent
#define	RESOLV_CONF	"/etc/resolv.conf"
#define	SETPWENT_TYPE	void
#define	SIGNAL_TYPE	void
#define	SPRINTF_TYPE	int
#define	STDARG_H	<stdarg.h>
#define	STDLIB_H	<stdlib.h>
#define	STRING_H	<string.h>
#define	STRLEN_TYPE	size_t
#define	SYS_STREAM_H	<sys/stream.h>
#define	TIMES_TYPE	clock_t
#define	TIME_H	<sys/time.h>
#define	UMASK_TYPE	mode_t
#define	UNISTD_H	<unistd.h>
#define	USE_BSDNODELAY	
#define	USE_DUP2	
#define	USE_FCHMOD	
#define	USE_FCHOWN	
#define	USE_FCNTL
#define	USE_FREAD
#define	USE_FSYNC	
#define	USE_FTRUNCATE	
#define	USE_FWRITE
#define	USE_GETEUID
#define	USE_GETPWNAM		# define this if getpwent() iteration is inappropriate
#define	USE_GETUID
#define	USE_INET
#define	USE_LIB
#define	USE_LSTAT	
#define	USE_MAILLOCK	
#define	USE_MAILLOCK
#define	USE_MALLOC
#define	USE_MEMMOVE	
#define	USE_MNTENT
#define	USE_NDBM
#define	USE_NDIR
#define	USE_NFS
#define	USE_RENAME	
#define	USE_RESOLV
#define	USE_SELECT	
#define	USE_SETEUID	
#define	USE_SETPWENT
#define	USE_SETSID	
#define	USE_SETVBUF	
#define	USE_SIGNAL
#define	USE_SPRINTF
#define	USE_STATBLKSIZ	
#define	USE_STDARG
#define	USE_STDLIB
#define	USE_STRCHR	
#define	USE_STRCSPN	
#define	USE_STRERROR	
#define	USE_STRING
#define	USE_STRLEN
#define	USE_STRSIGNAL	
#define	USE_SYS
#define	USE_SYSLOG
#define	USE_SYSVTIMEZONE	
#define	USE_TIME
#define	USE_TIMES
#define	USE_TIMEVAL	
#define	USE_UMASK
#define	USE_UNAME	
#define	USE_UNISTD
#define	USE_VFPRINTF	
#define	gethostbyaddr	res_gethostbyaddr
#define	gethostbyname	res_gethostbyname
