/*
 *	Copyright 1990 by Rayan S. Zachariassen, all rights reserved.
 *	This will be free software, but only when it is finished.
 */

#include "hostenv.h"

#ifndef	USE_STRSIGNAL
char *
strsignal(sig)
	int sig;
{
	static char buf[64];

#ifdef USE_SIGLIST
	extern char *sys_siglist[];

	if ((unsigned)sig < NSIG)
		return sys_siglist[sig];
#endif /* USE_SYS_SIGLIST */
	sprintf(buf, "Signal %d", sig);
	return buf;
}
#endif /* !USE_STRSIGNAL */
